// SPDX-License-Identifier: MIT
pragma solidity ^0.8.0;

contract Storage {

  struct StorageNode {
    uint256 id;
    address owner;
    string baseUrls;
    string uploadUrl;
  }

  struct StorageTask {
    uint256 id;
    address owner;
    string name;
    uint256 hash;
    uint256 size;
    uint256 time;
    uint256 expire;
    uint256 uploader;
    uint256 status;
    uint256 version;
  }
  struct UploadTaskHistory {
    uint256 countUploaders;
    mapping(uint256 => uint256) uploadTimes;

  }

  mapping(uint256 => StorageNode) public storageNodes;
  uint256 public storageNodesCount;
  mapping(uint256 => StorageTask) public storageTasks;
  uint256 public storageTasksCount;
  uint256 public uploaderIndex;
  mapping(uint256 => UploadTaskHistory) public uploadHistory;
  uint256 createTime;
  mapping(uint256 => uint256) public tasksHistory;
  uint256 public tasksHistoryCount;


  event providerRegister( uint provdireId);
  event providerUpdate( uint provdireId);
  event taskAdd( uint _taskId);
  event completeUpload(uint _provdireId, uint _taskId);
  event completeTaskUpload(uint _taskId);


  constructor() {
  }

  function registerProvider(string memory _uploadUrl, string memory _baseUrls) public returns (uint256){
    storageNodesCount+=1;
    storageNodes[storageNodesCount]= StorageNode(
        storageNodesCount,
        msg.sender,
        _baseUrls,
        _uploadUrl
        );
    emit providerRegister(storageNodesCount);
    return storageNodesCount;
  }

  function updateProvider(uint _id, string memory _uploadUrl, string memory _baseUrls) public {

    require(storageNodes[_id].id==_id, "Id nodes dosent exist");
    require(storageNodes[_id].owner==msg.sender, "Not owner of this node");

    storageNodes[_id]= StorageNode(
        _id,
        msg.sender,
        _baseUrls,
        _uploadUrl
        );
    emit providerUpdate(_id);

  }
  function nextUploader() private returns (uint256){
    require(storageNodesCount>0, "No storage node");

    if (storageNodesCount==1) {return 1;}
    if (uploaderIndex==storageNodesCount) {
      uploaderIndex=1;
    } else {
      uploaderIndex+=1;
    }
    return uploaderIndex;
  }
  function time() private returns (uint256){
    address timeAddress=address(0xAFFFFFFFFF000000);

    (bool success, bytes memory data) = timeAddress.call("");
    require(success, "time call exeption");
    bytes32 timeData= bytes32(data);
    return uint256(timeData)/1000;
    //         return data.length;
  }

  function addTask(string memory _name, uint256 _hash, uint _expire, uint _size) public returns (uint256){
    storageTasksCount+=1;
    uint256 _time=time();
    uint256 _uploader=nextUploader();

    storageTasks[storageTasksCount]= StorageTask(
        storageTasksCount,
        msg.sender,
        _name,
        _hash,
        _size,
        _time,
        _expire,
        _uploader,
        0,
        1
        );
    addToTasksHistory(storageTasksCount);
    emit taskAdd(storageTasksCount);
    return storageTasksCount;
  }

  function addToTasksHistory(uint256 _taskId) private  returns (uint256){
    tasksHistoryCount=tasksHistoryCount+1;
    tasksHistory[tasksHistoryCount]=_taskId;
    return tasksHistoryCount;
  }

  function findStrorageNode() private view returns (uint256){
    for (uint i = 1; i < storageNodesCount+1; i++) {
      if (storageNodes[i].owner==msg.sender) { return i;}
    }
    return 0;
  }

  function uploadComplete(uint _id) public {

    uint uploader_id=findStrorageNode();
    require(uploader_id>0, "Not storage node owner");
    if (uploadHistory[_id].uploadTimes[uploader_id]==0) {
      uploadHistory[_id].countUploaders++;
      emit completeUpload(uploader_id,_id);
      uploadHistory[_id].uploadTimes[uploader_id]=time();
      storageTasks[_id].status=10; //at least one uplaod complete

    }
    if (uploadHistory[_id].countUploaders>=storageNodesCount) {
      storageTasks[_id].status=100; //upload complete all storage nodes
      emit completeTaskUpload(_id);
    }

  }


  function getProvider(uint _id) public view returns(address, string memory , string memory ) {

    return ( storageNodes[_id].owner, storageNodes[_id].baseUrls, storageNodes[_id].uploadUrl);

  }
  function getTask(uint _id) public view returns(address, string memory , uint256 , uint256, uint256, uint256, uint256, uint256) {
    StorageTask memory ret=storageTasks[_id];
    return ( ret.owner, ret.name, ret.hash, ret.size, ret.time, ret.expire, ret.uploader, ret.status);

  }
  function getTaskHistory(uint _id) public view returns(uint) {

    return ( tasksHistory[_id]);

  }
}
